from __future__ import unicode_literals
"""XML namespaces"""


admin = "http://moyaproject.com/admin"
auth = "http://moyaproject.com/auth"
blog = "http://moyaproject.com/blog"
comments = "http://moyaproject.com/comments"
db = "http://moyaproject.com/db"
default = "http://moyaproject.com"
email = "http://moyaproject.com/email"
feedback = "http://moyaproject.com/feedback"
forms = "http://moyaproject.com/forms"
fs = "http://moyaproject.com/fs"
html = "http://moyaproject.com/html"
image = "http://moyaproject.com/image"
jsonrpc = "http://moyaproject.com/jsonrpc"
let = "http://moyaproject.com/let"
links = "http://moyaproject.com/links"
preflight = "http://moyaproject.com/preflight"
recaptcha = "http://moyaproject.com/recaptcha"
run = default
tables = "http://moyaproject.com/tables"
test = "http://moyaproject.com/test"
thumbnail = "http://moyaproject.com/thumbnail"
widgets = "http://moyaproject.com/widgets"
wysihtml5 = "http://moyaproject.com/wysihtml5"


namespace_docs = {
    "http://moyaproject.com/admin": """
        Tags defined in the [link admin]Moya Admin[/link] library.

    """,
    "http://moyaproject.com/auth": """
        Tags defined in [link auth]Moya Auth[/link] library.
    """,
    "http://moyaproject.com/blog": """
        Tags defined in the [link blog]Moya Blog[/link] library.
    """,
    "http://moyaproject.com/comments": """
        Tags defined in the [link comments]Moya Comments[/link] library.
    """,
    "http://moyaproject.com/db": """
        Tags used to work with [link db]databases[/link].
    """,
    "http://moyaproject.com": """
        The default namespace used for most of Moya's tags.
    """,
    "http://moyaproject.com/email": """
        Tags related to [link email]email[/link].
    """,
    "http://moyaproject.com/feedback": """
        Tags defined in [link feedbac]Moya Feedback[/link].
    """,
    "http://moyaproject.com/forms": """
        Tags defined in [link forms]Moya Forms[/link].
    """,
    "http://moyaproject.com/fs": """
        Tags for working with [link project#filesystems]filesystems[/link].
    """,
    "http://moyaproject.com/image": """
        Tags for working with [link images]images[/link].
    """,
    "http://moyaproject.com/jsonrpc": """
        Tags for creating [link jsonrpc]JSON RPC[/link] interfaces.
    """,
    "http://moyaproject.com/links": """
        Tags for defining [link links]links[/link].
    """,
    "http://moyaproject.com/preflight": """
        Tags for creating [link preflight]preflight checks[/link].
    """,
    "http://moyaproject.com/recaptcha": """
        Tags defined in [link recaptcha]Moya Google Recaptcha[/link].
    """,
    "http://moyaproject.com/tables": """
        Tags used to create [link tables]table[/link] content.
    """,
    "http://moyaproject.com/test": """
        Tags to build unit tests.
    """,
    "http://moyaproject.com/thumbnail": """
        Tags defined in the [link thumbnail]Moya Thumbnail[/link] library.
    """,
    "http://moyaproject.com/widgets": """
        Widgets defined in [link widgets]Moya Widgets[/link].
    """,
    "http://moyaproject.com/wysihtml5": """
        Tags to create a rich text editor with [link wysihtml5]WYSIHTML5[/link].
    """
}
