<button type="submit" {% attrib name=field.name, value=field.clicked %} class="btn{% if field.class %} ${field.class}{% else %} btn-primary{% endif %}">{% if field.text %}${field.text}{% endif %}</button>
