{% extends "/base.html" %}

{% block "body" append %}

{%- if filename %}
{%- for thumb in ['small', 'square', 'width200', 'fitsquare', 'fit'] %}
<img src="${filename | "thumbnail_url"(processor=thumb)}">
{%- endfor %}
{%- endif %}

{%- endblock %}
