<div class="checkbox">
    <label>
        <input id="${field.id}" name="${field.name}" type="checkbox"
            {% if field.value %}checked="checked"{% endif %} value="${field.on}"/> ${field.text}
    </label>
</div>